"use strict";

var _classCallCheck = require("babel-runtime/helpers/class-call-check")["default"];

var _inherits = require("babel-runtime/helpers/inherits")["default"];

var _core = require("babel-runtime/core-js")["default"];

var _interopRequire = require("babel-runtime/helpers/interop-require")["default"];

var Visitor = _interopRequire(require("./visitor"));

var Parser = _interopRequire(require("message-format/parser"));

var Linter = (function (_Visitor) {
  function Linter(options) {
    _classCallCheck(this, Linter);

    _Visitor.call(this, options);
  }

  _inherits(Linter, _Visitor);

  Linter.prototype.lint = function lint(_ref) {
    var sourceCode = _ref.sourceCode;
    var sourceFileName = _ref.sourceFileName;

    this.run({ sourceCode: sourceCode, sourceFileName: sourceFileName });
  };

  Linter.prototype.visitFormatCall = function visitFormatCall(path, traverser, opts) {
    var _this = this;

    traverser.traverse(path); // pre-travserse children

    var node = path.node;
    var isTranslateOnly = opts.isTranslateOnly;

    var numErrors = 0;

    var patternNode = node.arguments[0];
    var patternIsString = this.isString(patternNode);
    var pattern = patternIsString && this.getStringValue(patternNode);
    var patternError = pattern && this.getPatternError(pattern);
    var patternAst = pattern && !patternError && Parser.parse(pattern);
    var patternParams = patternAst && this.getPatternParams(patternAst);

    var argsNode = isTranslateOnly ? null : node.arguments[1];
    var argsIsLiteral = this.isObjectLiteral(argsNode);
    var argsPassed = argsIsLiteral && this.getObjectKeys(argsNode);

    var localeNode = isTranslateOnly ? node.arguments[1] : node.arguments[2];
    var localeIsString = this.isString(localeNode);
    var locale = localeIsString && this.getStringValue(localeNode);

    if (!patternIsString) {
      this.reportWarning(path, "Warning: called without a literal pattern");
      ++numErrors;
    }

    if (patternError) {
      this.reportError(path, "SyntaxError: pattern is invalid", patternError.message);
      ++numErrors;
    }

    if (!isTranslateOnly && patternParams && patternParams.length > 0 && !argsNode) {
      this.reportError(path, "TypeError: pattern requires parameters, but called with no arguments");
      ++numErrors;
    }

    if (!isTranslateOnly && patternParams && patternParams.length > 0 && argsIsLiteral) {
      var missingArgs = patternParams.filter(function (param) {
        return argsPassed.indexOf(param) === -1;
      });
      missingArgs.forEach(function (arg) {
        _this.reportError(path, "TypeError: pattern requires parameter \"" + arg + "\", but it is missing from the arguments object");
        ++numErrors;
      });
    }

    if (this.isReplaceable(path) && this.translations) {
      var translation = this.translate(pattern, locale || this.locale);
      var translationError = translation && this.getPatternError(translation);
      var translationAst = translation && !translationError && Parser.parse(translation);
      var translationParams = translationAst && this.getPatternParams(translationAst);

      if (translation == null) {
        this.reportWarning(path, "Warning: no " + (locale || this.locale) + " translation found for key " + JSON.stringify(this.getKey(pattern)));
        ++numErrors;
      }

      if (translationError) {
        this.reportError(path, "SyntaxError: " + (locale || this.locale) + " translated pattern is invalid for key " + JSON.stringify(this.getKey(pattern)), translationError.message);
        ++numErrors;
      }

      if (!isTranslateOnly && translationParams && translationParams.length > 0 && !argsNode) {
        this.reportError(path, "TypeError: " + (locale || this.locale) + " translated pattern requires parameters, but called with no arguments");
        ++numErrors;
      }

      if (!isTranslateOnly && translationParams && translationParams.length > 0 && argsIsLiteral) {
        var missingArgs = translationParams.filter(function (param) {
          return argsPassed.indexOf(param) === -1;
        });
        missingArgs.forEach(function (arg) {
          _this.reportError(path, "TypeError: " + (locale || _this.locale) + " translated pattern requires parameter \"" + arg + "\", but it is missing from the arguments object");
          ++numErrors;
        });
      }
    }

    if (localeNode && !localeIsString) {
      this.reportWarning(path, "Warning: called with a non-literal locale");
      ++numErrors;
    }

    return numErrors;
  };

  Linter.prototype.isObjectLiteral = function isObjectLiteral(node) {
    return node && node.type === "ObjectExpression";
  };

  Linter.prototype.getObjectKeys = function getObjectKeys(object) {
    return object.properties.filter(function (prop) {
      return prop.key.type === "Identifier";
    }).map(function (prop) {
      return prop.key.name;
    });
  };

  Linter.prototype.getPatternParams = function getPatternParams(patternAst) {
    var _this = this;

    return patternAst.reduce(function (list, element) {
      if (typeof element === "string") return list;
      var type = element[1];
      if (type === "select" || type === "plural" || type === "selectordinal") {
        (function () {
          var newList = list.concat(element[0]);
          var children = type === "select" ? element[2] : element[3];
          _core.Object.keys(children).forEach(function (key) {
            newList = newList.concat(_this.getPatternParams(children[key]));
          });
        })();
      } else {
        return list.concat(element[0]);
      }
    }, []);
  };

  Linter.lint = function lint(source, options) {
    return new Linter(options).lint(source);
  };

  Linter.lintFiles = function lintFiles(files, options) {
    var linter = new Linter(options);
    linter.forEachFile(files, function (source) {
      return linter.lint(source);
    });
  };

  return Linter;
})(Visitor);

module.exports = Linter;