"use strict";

var _classCallCheck = require("babel-runtime/helpers/class-call-check")["default"];

var _core = require("babel-runtime/core-js")["default"];

var _interopRequire = require("babel-runtime/helpers/interop-require")["default"];

var locales = _interopRequire(require("./locales.json"));

var lookupClosestLocale = _interopRequire(require("message-format/lib/lookup-closest-locale"));

var formats = require("message-format/lib/data").formats;

var getKeyUnderscoredCrc32 = require("./translate-util").getKeyUnderscoredCrc32;

/**
 * Transpiler
 *
 * Turns this:
 *  [ "You have ", [ "numBananas", "plural", 0, {
 *       "=0": [ "no bananas" ],
 *      "one": [ "a banana" ],
 *    "other": [ [ '#' ], " bananas" ]
 *  } ], " for sale." ]
 *
 * into this:
 *  `(function(args, locale) {
 *    return "You have " + formatMessage.plural(locale, args["numBananas"], 0, {
 *      "=0": "no bananas",
 *      "one": "a banana",
 *      "other": function() {
 *        return args["numBananas"] + " bananas";
 *      }
 *    }) + " for sale.";
 *  })(args, "en")`
 **/

var Transpiler = (function () {
  function Transpiler() {
    var options = arguments[0] === undefined ? {} : arguments[0];

    _classCallCheck(this, Transpiler);

    this.vars = {};
    this.locale = options.locale || "en";
    this.functionName = options.functionName || "formatMessage";
    this.paramsNode = options.paramsNode;
    this.originalPattern = options.originalPattern;
  }

  Transpiler.prototype.checkUseArgExpressions = function checkUseArgExpressions(elements) {
    var blacklist = ["plural", "selectordinal", "select"];
    var noIntermediateVars = elements.every(function (element) {
      return typeof element === "string" || blacklist.indexOf(element[1]) === -1;
    });
    var noPossibleSideEffects = this.paramsNode && (this.paramsNode.type === "ObjectExpression" && this.paramsNode.properties.every(function (prop) {
      return prop.key.type === "Identifier" && (prop.value.type === "Identifier" || prop.value.type === "Literal");
    }));
    this.useArgExpressions = false;
    if (noIntermediateVars && noPossibleSideEffects) {
      this.useArgExpressions = this.paramsNode.properties.reduce(function (argMap, prop) {
        argMap[prop.key.name] = prop.value.raw || prop.value.name;
        return argMap;
      }, {});
    }
  };

  Transpiler.prototype.transpile = function transpile(elements) {
    var _this = this;

    this.vars = {};

    if (elements.length === 0) {
      return { replacement: "\"\"" };
    }

    if (elements.length === 1 && typeof elements[0] === "string") {
      return { replacement: JSON.stringify(elements[0]) };
    }

    this.checkUseArgExpressions(elements);

    elements = elements.map(function (element) {
      return _this.transpileElement(element, null);
    });

    if (this.useArgExpressions) {
      return { replacement: elements.join(" + ") };
    }

    var vars = _core.Object.keys(this.vars);
    var key = this.originalPattern || elements.join(" ");
    var functionName = "$$_" + getKeyUnderscoredCrc32(key);
    var replacement = functionName + "(" + this.functionName + ", args)"; // args needs to be swapped by consumer
    var declaration = "function " + functionName + " (" + this.functionName + ", args) {\n" + (!vars.length ? "" : "  var " + vars.join(", ") + ";\n") + "  return " + elements.join(" + ") + ";\n}";
    return { replacement: replacement, declaration: declaration };
  };

  Transpiler.prototype.transpileSub = function transpileSub(elements, parent) {
    var _this = this;

    if (elements.length === 0) {
      return "\"\"";
    }

    if (elements.length === 1 && typeof elements[0] === "string") {
      return JSON.stringify(elements[0]);
    }

    elements = elements.map(function (element) {
      return _this.transpileElement(element, parent);
    });

    return "(" + elements.join(" + ") + ")";
  };

  Transpiler.prototype.transpileElement = function transpileElement(element, parent) {
    if (typeof element === "string") {
      return JSON.stringify(element);
    }

    var id = element[0];
    var type = element[1];
    var style = element[2];

    if (id === "#") {
      id = parent[0];
      var _offset = parent[2];
      return this.transpileNumber(id, _offset, null);
    }

    switch (type) {
      case "number":
      case "ordinal": // TODO: rbnf
      case "spellout": // TODO: rbnf
      case "duration":
        // TODO: duration
        return this.transpileNumber(id, 0, style);
      case "date":
      case "time":
        return this.transpileDateTime(id, type, style);
      case "plural":
      case "selectordinal":
        var offset = element[2];
        var options = element[3];
        return this.transpilePlural(id, type, offset, options);
      case "select":
        return this.transpileSelect(id, style);
      default:
        return this.transpileArgument(id);
    }
  };

  Transpiler.prototype.transpileNumber = function transpileNumber(id, offset, style) {
    return this.functionName + ".number(" + JSON.stringify(this.locale) + ", " + this.transpileArgument(id) + (offset ? "-" + offset : "") + (style ? ", " + JSON.stringify(style) : "") + ")";
  };

  Transpiler.prototype.transpileDateTime = function transpileDateTime(id, type, style) {
    return this.functionName + "." + type + "(" + JSON.stringify(this.locale) + ", " + this.transpileArgument(id) + (style ? ", " + JSON.stringify(style) : "") + ")";
  };

  Transpiler.prototype.transpilePlural = function transpilePlural(id, type, offset, children) {
    var _this = this;

    var parent = [id, type, offset /*, children*/];
    var closest = lookupClosestLocale(this.locale, locales.locales);
    var conditions = locales.locales[closest].plurals[type === "selectordinal" ? "ordinal" : "cardinal"];
    var cond = undefined;
    var other = "\"\"";
    var select = "";
    var exact = "";
    var sub = undefined;
    var vars = ["s=" + this.transpileArgument(id), "n=+s"];
    var pvars = [];
    var refsI = false;
    var refsV = false;
    var refsW = false;
    var refsF = false;
    var refsT = false;
    this.vars.s = true;
    this.vars.n = true;
    if (offset) {
      pvars.push("s=+s-" + offset);
      pvars.push("n=s");
    }

    _core.Object.keys(children).forEach(function (key, i) {
      sub = "\n      " + _this.transpileSub(children[key], parent).replace(/\n/g, "\n      ");
      if (key === "other") {
        other = sub;
      } else if (key.charAt(0) === "=") {
        exact += "\n    " + key.slice(1) + " === n ?" + sub + " :";
      } else if (key in conditions) {
        cond = conditions[key];
        if (/\bi\b/.test(cond)) {
          _this.vars.i = refsI = true;
        }
        if (/\bv\b/.test(cond)) {
          _this.vars.v = refsV = true;
        }
        if (/\bw\b/.test(cond)) {
          _this.vars.w = refsW = true;
        }
        if (/\bf\b/.test(cond)) {
          _this.vars.f = refsF = true;
        }
        if (/\bt\b/.test(cond)) {
          _this.vars.t = refsT = true;
        }
        select += "\n    /*" + key + "*/(" + cond + ") ?" + sub + " :";
      }
    });
    if (refsI) {
      pvars.push("i=" + locales.pluralVars.i);
    }
    if (refsV) {
      pvars.push("v=" + locales.pluralVars.v);
    }
    if (refsW) {
      pvars.push("w=" + locales.pluralVars.w);
    }
    if (refsF) {
      pvars.push("f=" + locales.pluralVars.f);
    }
    if (refsT) {
      pvars.push("t=" + locales.pluralVars.t);
    }
    return "(\n    (" + vars.join(", ") + "," + exact + (pvars.length ? "\n\n    (" + pvars.join(", ") + "," + select + other + ")\n    ))" : select + other + ")\n    )");
  };

  Transpiler.prototype.transpileSelect = function transpileSelect(id, children) {
    var _this = this;

    var other = "\"\"";
    var select = "(\n    (s=" + this.transpileArgument(id) + ",";
    this.vars.s = true;
    _core.Object.keys(children).forEach(function (key, i) {
      if (key === "other") {
        other = "\n      " + _this.transpileSub(children[key]).replace(/\n/g, "\n      ");
        return;
      }
      select += "\n    " + JSON.stringify(key) + " === s ?\n      " + _this.transpileSub(children[key]).replace(/\n/g, "\n      ") + " :";
    });
    select += other + ")\n    )";
    return select;
  };

  Transpiler.prototype.transpileArgument = function transpileArgument(id) {
    return this.useArgExpressions ? this.useArgExpressions[id] : "args[" + JSON.stringify(id) + "]";
  };

  Transpiler.transpile = function transpile(elements, options) {
    return new Transpiler(options).transpile(elements);
  };

  return Transpiler;
})();

module.exports = Transpiler;