"use strict";

var _classCallCheck = require("babel-runtime/helpers/class-call-check")["default"];

var _core = require("babel-runtime/core-js")["default"];

var _interopRequire = require("babel-runtime/helpers/interop-require")["default"];

var _fs = require("fs");

var readFileSync = _fs.readFileSync;
var writeFileSync = _fs.writeFileSync;

var _path = require("path");

var relative = _path.relative;
var resolve = _path.resolve;
var dirname = _path.dirname;
var basename = _path.basename;
var pathJoin = _path.join;

var mkdirp = _interopRequire(require("mkdirp"));

var recast = _interopRequire(require("recast"));

var sourceMap = _interopRequire(require("source-map"));

var chalk = _interopRequire(require("chalk"));

var Parser = _interopRequire(require("message-format/parser"));

var _translateUtil = require("./translate-util");

var getTranslate = _translateUtil.getTranslate;
var getGetKey = _translateUtil.getGetKey;
var getKeyNormalized = _translateUtil.getKeyNormalized;
var getKeyUnderscoredCrc32 = _translateUtil.getKeyUnderscoredCrc32;

var Transpiler = _interopRequire(require("./transpiler"));

var packageJson = _interopRequire(require("../package.json"));

var Literal = recast.types.namedTypes.Literal.toString();
var TemplateLiteral = recast.types.namedTypes.TemplateLiteral.toString();
var CallExpression = recast.types.namedTypes.CallExpression.toString();
var ImportDefaultSpecifier = recast.types.namedTypes.ImportDefaultSpecifier.toString();

/**
 * Base class for traversing source code to lint, extract, and inline messages
 **/

var Visitor = (function () {
  function Visitor() {
    var options = arguments[0] === undefined ? {} : arguments[0];

    _classCallCheck(this, Visitor);

    this.functionName = options.functionName || "formatMessage";
    this.autoDetectFunctionName = "autoDetectFunctionName" in options ? !!options.autoDetectFunctionName : true;
    this.getKey = getGetKey(options);
    this.translate = getTranslate(options);
    this.translations = options.translations;
    this.missingTranslation = options.missingTranslation;
    this.missingReplacement = options.missingReplacement;
    this.locale = options.locale || "en";
    this.currentFileName = null;

    if (typeof options.emitWarning === "function") {
      this.emitWarning = options.emitWarning;
    }
    if (typeof options.emitError === "function") {
      this.emitError = options.emitError;
    }
    if (typeof options.emitResult === "function") {
      this.emitResult = options.emitResult;
    }
    if (typeof options.emitFile === "function") {
      this.emitFile = options.emitFile;
    }
  }

  Visitor.prototype.emitWarning = function emitWarning(message) {
    console.warn(message);
  };

  Visitor.prototype.emitError = function emitError(message) {
    console.error(message);
  };

  Visitor.prototype.emitResult = function emitResult(code) {
    console.log(code);
  };

  Visitor.prototype.emitFile = function emitFile(filename, content, sourcemap) {
    mkdirp.sync(dirname(filename));
    writeFileSync(filename, content, "utf8");
  };

  Visitor.prototype.reportWarning = function reportWarning(path, message) {
    this.emitWarning(chalk.bold.yellow(message) + "\n" + chalk.yellow(this.getLocation(path)));
  };

  Visitor.prototype.reportError = function reportError(path, message, submessage) {
    this.emitError(chalk.red(chalk.bold(message) + "\n" + this.getLocation(path) + (!submessage ? "" : "\n    " + submessage.replace(/\n/g, "\n    "))));
  };

  Visitor.prototype.getLocation = function getLocation(path) {
    var translate = this.isTranslateCall(path) ? ".translate" : "";

    return "    at " + this.functionName + translate + " (" + this.currentFileName + ":" + path.node.loc.start.line + ":" + path.node.loc.start.column + ")";
  };

  Visitor.prototype.isFormatImport = function isFormatImport(node) {
    return node && (node.source && node.source.value === packageJson.name && node.specifiers && node.specifiers[0] && node.specifiers[0].type === ImportDefaultSpecifier);
  };

  Visitor.prototype.isFormatRequire = function isFormatRequire(node) {
    var depth = arguments[1] === undefined ? 0 : arguments[1];

    return node && (node.type === CallExpression && node.callee.name === "require" && this.isString(node.arguments[0]) && this.getStringValue(node.arguments[0]) === packageJson.name) || node && (
    // _interopRequire or other wrapper
    node.type === CallExpression && depth <= 1 // allow at most one wrapper around require
     && this.isFormatRequire(node.arguments[0], depth + 1));
  };

  Visitor.prototype.isFormatCall = function isFormatCall(path) {
    var node = path.node;
    return node.callee && node.callee.name === this.functionName;
  };

  Visitor.prototype.isTranslateCall = function isTranslateCall(path) {
    var node = path.node;
    var callee = node.callee;
    return callee && (
    // member expression
    callee.object && callee.object.name === this.functionName && callee.property && callee.property.name === "translate");
  };

  Visitor.prototype.isReplaceable = function isReplaceable(path) {
    var node = path.node;
    return node && (this.isFormatCall(path)
    // first argument is a literal string, or template literal with no expressions
     && this.isString(node.arguments[0]) && (
    // no specified locale, or is a literal string
    !node.arguments[2] || this.isString(node.arguments[2])) || this.isTranslateCall(path) && this.isString(node.arguments[0]) && (!node.arguments[1] || this.isString(node.arguments[1])));
  };

  Visitor.prototype.isString = function isString(node) {
    return node && (node.type === Literal && typeof node.value === "string" || node.type === TemplateLiteral && node.expressions.length === 0 && node.quasis.length === 1);
  };

  Visitor.prototype.getStringValue = function getStringValue(literal) {
    if (literal.type === TemplateLiteral) {
      return literal.quasis[0].value.cooked;
    }
    return literal.value;
  };

  Visitor.prototype.getPatternError = function getPatternError(pattern) {
    try {
      Parser.parse(pattern);
    } catch (error) {
      return error;
    }
  };

  Visitor.prototype.forEachFile = function forEachFile(files, fn, ctx) {
    for (var _iterator = files, _isArray = Array.isArray(_iterator), _i = 0, _iterator = _isArray ? _iterator : _core.$for.getIterator(_iterator);;) {
      var _ref;

      if (_isArray) {
        if (_i >= _iterator.length) break;
        _ref = _iterator[_i++];
      } else {
        _i = _iterator.next();
        if (_i.done) break;
        _ref = _i.value;
      }

      var file = _ref;

      var source = file;
      if (typeof file === "string") {
        source = {
          sourceFileName: file,
          sourceCode: readFileSync(file, "utf8")
        };
      }
      fn.call(ctx, source);
    }
  };

  Visitor.prototype.run = function run(_ref) {
    var _this = this;

    var sourceCode = _ref.sourceCode;
    var sourceFileName = _ref.sourceFileName;

    var ast = this.parse({ sourceCode: sourceCode, sourceFileName: sourceFileName });
    if (ast) {
      (function () {
        var self = _this;
        recast.visit(ast, _this.visitors({
          visitCallExpression: function visitCallExpression(path) {
            var isTranslateOnly = self.isTranslateCall(path);
            if (isTranslateOnly || self.isFormatCall(path)) {
              self.visitFormatCall(path, this, { isTranslateOnly: isTranslateOnly });
            } else {
              this.traverse(path);
            }
          }
        }));
      })();
    }
    return ast;
  };

  Visitor.prototype.parse = function parse(_ref) {
    var sourceCode = _ref.sourceCode;
    var sourceFileName = _ref.sourceFileName;

    this.currentFileName = sourceFileName;
    try {
      return recast.parse(sourceCode, { sourceFileName: sourceFileName });
    } catch (error) {
      this.emitError(chalk.red(chalk.bold("SyntaxError: Could not parse source code") + "\n" + "    at " + sourceFileName + "\n" + "    " + error.message));
    }
  };

  Visitor.prototype.print = function print(_ref) {
    var ast = _ref.ast;
    var sourceMapName = _ref.sourceMapName;
    var inputSourceMap = _ref.inputSourceMap;

    return recast.print(ast, { sourceMapName: sourceMapName, inputSourceMap: inputSourceMap });
  };

  Visitor.prototype.visitors = function visitors(base) {
    var _this = this;

    if (this.autoDetectFunctionName) {
      (function () {
        var self = _this;
        _core.Object.assign(base, {
          // simplistic function and file scope tracking
          visitFunction: function visitFunction(path) {
            var functionName = self.functionName;
            this.traverse(path); // traverse children
            self.functionName = functionName;
          },

          visitProgram: function visitProgram(path) {
            var functionName = self.functionName;
            this.traverse(path); // traverse children
            self.functionName = functionName;
          },

          visitVariableDeclarator: function visitVariableDeclarator(path) {
            this.traverse(path); // pre-travserse children
            if (self.isFormatRequire(path.node.init)) {
              self.functionName = path.node.id.name;
            }
          },

          visitImportDeclaration: function visitImportDeclaration(path) {
            this.traverse(path); // pre-travserse children
            if (self.isFormatImport(path.node)) {
              self.functionName = path.node.specifiers[0].id.name;
            }
          }
        });
      })();
    }
    return base;
  };

  return Visitor;
})();

module.exports = Visitor;